#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "general.h"
#include "util.h"
#include "viz.h"
#include "nmatrix.h"
#include "matrix.h"
#include "file.h"
#include "bonds.h"

/************************************************************************/
/*  defines																*/
/************************************************************************/

/************************************************************************/
/*  typedefs															*/
/************************************************************************/

/************************************************************************/
/*  function prototypes													*/
/************************************************************************/
/*
 * File menu callbacks
 */
void load_db( void *data );
void sim_db( void *data, int dims, int count, int clust );
void report_db( void *data );
void quit( void *data );

int init_files( DB *db );
void report_config( DB *db );
int simulate_data( DB *db, int dims, int count, int num_clusters );

char *read_quoted_string( char *buf, char *str );
char *read_string( char *buf, char *str );
int add_nominal( DATUM *datum, long j, char *tok );
 
/************************************************************************/
/*  external functions 													*/
/************************************************************************/
extern char *GetFile( char * );
extern void free_data( DB *db );
extern void setup_screen_coords( DB *db );
//extern void set_mode_toggles( int mode );

/************************************************************************/
/*  global variables													*/
/************************************************************************/
extern char errbuf[];

/************************************************************************/
/*  Function:	init_files - opens any files specified by the user		*/
/*				on the command line										*/
/*																		*/
/*	Inputs:																*/
/*		db		- void pointer to database								*/
/*																		*/
/*	Returns:															*/
/*		none															*/
/************************************************************************/
int init_files( DB *db )
{
	DEBUG( "init_files" );

	if( (db->files.reportfp = fopen( db->files.outfile, "w" )) == NULL )
		return( ERROR );

	return( SUCCESS );
}

/************************************************************************/
/*  Function:	load_db - callback for File menu load option. Prompts 	*/
/*				user for name of data file to read in.	Frees any old 	*/
/*				data if it exists and then	allocates space for the 	*/
/*				data and attempts to read it in							*/
/*																		*/
/*				This is one of the two ways data can get into 			*/
/*				MAVIS. So, when the data is read in we normalize each	*/
/*				dimension, calculate all the n-dimensional distances	*/
/*				and set up screen coordinates. Then we draw the data to	*/
/*				give the user feedback that the data has beed read in	*/
/*																		*/
/*	Inputs:																*/
/*		Widget	- integer id of widget that initiated callback			*/
/*		data	- void pointer to database								*/
/*																		*/
/*	Returns:															*/
/*		none															*/
/************************************************************************/
void load_db( void *data )
{
	FILE	*fp;
	DB		*db = (DB *)data;
	MATRIX	*D;
	
	DEBUG( "load_db" );

	/*
	 * open database file
	 */
	if( !db )
		return;

	if( (fp = fopen( db->files.infile, "r" )) == NULL )
	{
		sprintf(errbuf,"Error openning input data file: %s", db->files.infile);
		print_error( errbuf );
		return;
	}
		
	/*
	 * free old data if it exists
	 */
	if( db->dat.data )
		free_data( db );

	/*
	 * read database file
	 */
	if( read_db( fp, db ) == ERROR )
	{
		sprintf( errbuf,"Error reading input data file: %s", db->files.infile);
		print_error( errbuf );
		free_db( db );
		fclose( fp );
		return;
	}

	fclose( fp );

	/*
	 * save unnormalized data before we normalize
	 */
	copy_matrix( db->dat.R, db->dat.A );
	normalize_matrix( db->dat.A );

	/*
	 * now that we have data - calculate
	 * N-dimensional distances - ONE TIME ONLY :-)
	 */
	calc_all_distances( db );

#ifdef PDEBUG
	print_matrix( db->dat.A, "A" );
	print_matrix( db->dat.nD, "nD" );
	print_matrix( db->dat.O, "O" );
#endif

	/*
     * calculate screen coords after
     * clusters have been computed
	 * and draw db to show user
     */
    setup_screen_coords( db );
	/*calc_all_positions( db );*/

//	draw_db( db );
}

/****************************************************************/
/*	function:	setup_screen_coords() - initializes the display	*/
/*				coordinates for each data point in the database	*/
/*				It does this by finding which dimensions in the	*/
/*				database have the greatest variance, and then	*/
/*				scaling those dimensions to the width &height	*/
/*				of the screen to compute the screen coordinates	*/
/*																*/
/*	input:		db 		- pointer to database					*/
/*																*/
/*	output:		none											*/	
/****************************************************************/
void setup_screen_coords( DB *db )
{
	int		dims[MAX_DISPLAY_DIMS];
	long	i, j, width;
	long	k, active;
	MATRIX	*P;
	MATRIX	*PO;
	
	if( !db )
		return;

	P = db->dat.P;
	PO = db->dat.PO;
	
	switch( db->mds.display_init_mode )
	{
/*		case VARIANCE:
			/*
			 * find dimensions with greatest and 
			 * second greatest amounts of variance
			 * /
			find_axes( db, dims );
				
			/*
			 * use dimension with greatest variance to derive 
			 * x coordinates, and dimension with second greatest 
			 * variance to derive y coordinates 
			 * /
			for( i = 0; i < P->height ; i++ )
			{
			  	for( j = 0; j < P->width; j++ )
				{
					P->data[i][j] = db->dat.A->data[i][dims[j]];
					PO->data[i][j] =  P->data[i][j];
				}
			}
			break;
*/			
		case RANDOM:
			for( i = 0; i < P->height ; i++ )
			{
			  	for( j = 0; j < P->width; j++ )
				{
					P->data[i][j] = (nrand() - 0.5);
					PO->data[i][j] =  P->data[i][j];
				}
			}
			break;

		case DATA:
			/*
			 * derive display coords directly from data coords
			 */
			active = 0;
			for( i = 0; i < db->dat.A->width; i++ )
				active += db->dat.dim_mask[i];

			width = MIN( active, P->width );
			for( i = 0; i < P->height ; i++ )
			{
			  	for( k = 0, j = 0; j < width; j++, k++ )
				{
					for( ; !db->dat.dim_mask[k]; k++ )
						; /* empty loop body */
					
					P->data[i][j] = db->dat.A->data[i][k];
					PO->data[i][j] =  P->data[i][j];
				}
			}
			break;
			
		default:
			break;
			
	}
}

/************************************************************************/
/*  Function:	sim_db - callback for File menu simulate option. 		*/
/*				Opens dialog box to allow user to select what type of	*/
/*				simulated data to generate.	Frees any old data if it 	*/
/*				exists and then	allocates space for the data and 		*/
/*				attempts to fill in the data vectors					*/
/*																		*/
/*				This is one of the two ways data can get into 			*/
/*				MAVIS. So, when the data is read in we normalize each	*/
/*				dimension, calculate all the n-dimensional distances	*/
/*				and set up screen coordinates. Then we draw the data to	*/
/*				give the user feedback that the data has beed generated	*/
/*																		*/
/*	Inputs:																*/
/*		Widget	- integer id of widget that initiated callback			*/
/*		data	- void pointer to database								*/
/*																		*/
/*	Returns:															*/
/*		none															*/
/************************************************************************/
void sim_db( void *data, int dims, int count, int clust)
{
	DB		*db = (DB *)data;
	
	DEBUG( "sim_db" );

	/*
	 * open database file
	 */
	if( !db )
		return;

	if( simulate_data( db, dims, count, clust) == ERROR )
	{
		print_error( "Error generating simulated data" );
		free_db( db );
		return;
	}

	/*
	 * draw data to give user feedback
	 */
//	draw_db( db );
}

/************************************************************************/
/*  Function:	read_db - function that reads data from a file and		*/
/*				stores this data in database							*/
/*																		*/
/*	Inputs:																*/
/*		fp 	- FILE pointer to read data from							*/
/*		db	- pointer to databse to store data in						*/
/*																		*/
/*	Returns:															*/
/*		none															*/
/************************************************************************/
int read_db( FILE *fp, DB *db )
{
	long	i, j, n;
	long	num_points, data_dims;
	char	dim_type[MAX_NAME_LENGTH];
	char	dim_state[MAX_NAME_LENGTH];
	char	*ptr, *end, *tok;
	char	buf[MAX_NAME_LENGTH]; 
	char	label[MAX_NAME_LENGTH];
	MATRIX	*A;

	DEBUG( "read_db" );

	if( !db || read_line(fp, buf, MAX_NAME_LENGTH) == ERROR )
		return( ERROR );
	
	new_str( &db->name, buf );

	if( read_line(fp, buf, MAX_NAME_LENGTH) == ERROR )
		return( ERROR );
		
	/*
	 * read number of points and dimensions in database
	 */
	if( sscanf( buf,"%ld %ld", &db->dat.num_points, &db->dat.data_dims ) != 2 )
		return( ERROR );

	/*
	 * allocate data for database
	 */
	if( alloc_data( db ) == ERROR )
		return( ERROR );

	A = db->dat.A;

	for( i = 0; i < db->dat.data_dims; i++ )
	{
		if( read_line(fp, buf, MAX_NAME_LENGTH) == ERROR )
			return( ERROR );
			
		/*
		 * parse dimension info line
		 */
		if( (ptr = strchr(buf, '"')) == NULL )
		{
			if(!(ptr = read_string( buf, label )) )
				return( ERROR );
		}
		else
		{
			if(!(ptr = read_quoted_string( ptr, label )))
				return( ERROR );
		}

		db->dat.data_dim_array[i].label = mstrdup( label );

		if( sscanf( ptr, "%s %lf %lf %s",dim_type,
							&db->dat.data_dim_array[i].min,
							&db->dat.data_dim_array[i].max, dim_state ) != 4 )
		{
			return( ERROR );
		}

		tolower_string( dim_type );
		tolower_string( dim_state );

		/*
		 * convert dimension type to integer code
		 */
		if( streq( dim_type, "label" ) )
			db->dat.data_dim_array[i].type = LABEL;
		else if( streq( dim_type, "nominal" ) )
			db->dat.data_dim_array[i].type = NOMINAL;
		else if( streq( dim_type, "numeric" ) )
			db->dat.data_dim_array[i].type = NUMERIC;
		
		else
			return( ERROR );
	
		/*
		 * convert dimension enable/disable to mask value
		 */
		db->dat.dim_mask[i] = streq( dim_state, "disabled" ) ? 0 : 1;
	}
	
	/*
	 * read in actual data
	 */
	for( i = 0; i < db->dat.num_points; i++ )
	{
		if( read_line(fp, buf, MAX_NAME_LENGTH) == ERROR )
			return( ERROR );
			
		for(j = 0,tok = strtok(buf," \t\n");tok;j++,tok = strtok(NULL," \t\n"))
		{
			if( db->dat.data_dim_array[j].type == LABEL )
				db->dat.data[i].label = mstrdup( tok );
			else if( db->dat.data_dim_array[j].type == NOMINAL )
				add_nominal( &db->dat.data[i], j, tok );
			else
				A->data[i][j] = strtod(  tok, &end );
		}
		
		if( j != db->dat.data_dims )
			return( ERROR );
				
	}

	/*
	 * read in connection info
	 */
	for( i = 0; i < db->dat.num_points; i++ )
	{
		if( read_line(fp, buf, MAX_NAME_LENGTH) == ERROR )
			return( SUCCESS );

		tok = strtok(buf," \t\n");
		n = db->dat.data[i].num_connections = atoi( tok );
		
		for(j = 0; j < n; j++ )
		{
			tok = strtok(NULL," \t\n");
			db->dat.data[i].connections[j] = atoi( tok );
		}
	}
	
	return( SUCCESS );
	
}

int add_nominal( DATUM *datum, long j, char *tok )
{
	NOMINAL_INFO *new;

	DEBUG( "add_nominal" );

	if( (new = (NOMINAL_INFO *)mmalloc(sizeof(NOMINAL_INFO))) == NULL )
		return( ERROR );

	new->name = mstrdup( tok );
	new->column = j;
	new->next = NULL;

	if( !datum->nominal_head && !datum->nominal_tail )
	{
		datum->nominal_head = new;
		datum->nominal_tail = new;
	}
	else
	{
		datum->nominal_tail->next = new;
		datum->nominal_tail = new;
	}
}

/************************************************************************/
/*  Function:	read_quoted_string - reads a string with '"' marks and	*/
/*				the beginning and end - string can optionally contain	*/
/*				spaces													*/
/*																		*/
/*	Inputs:																*/
/*		buf	- character array to read quoted string from				*/
/*		str	- character array to return quoted string in				*/
/*																		*/
/*	Returns:															*/
/*		pointer to the remainder of the characters in buf, after str	*/
/*		is read															*/
/************************************************************************/
char *read_quoted_string( char *buf, char *str )
{
	char	*ptr;

	DEBUG( "read_quoted_string" );

	/*
	 * ending quote same as beginning quote
	 */
	if( (ptr = strrchr(buf, '"')) == buf )
		return( NULL );

	*ptr = '\0';
	strcpy( str, buf );

	return( ++ptr );
}

/************************************************************************/
/*  Function:	read_string - reads a string with no spaces in it		*/
/*																		*/
/*	Inputs:																*/
/*		buf	- character array to read string from						*/
/*		str	- character array to return string in						*/
/*																		*/
/*	Returns:															*/
/*		pointer to the remainder of the characters in buf, after str	*/
/*		is read															*/
/************************************************************************/
char *read_string( char *buf, char *str )
{
	char	*ptr, *ptr2;

	DEBUG( "read_string" );

	/*
	 * skip leading whitespace
	 */
	for( ptr = buf; isspace(*ptr); ptr++ )
		; /*empty */

	for( ptr2 = ptr; !isspace(*ptr2); ptr2++ )
		; /*empty */

	*ptr2 = '\0';
	strcpy( str, ptr );

	return( ++ptr2 );
}

/****************************************************************/
/*	function	simulate_data() - for every datum in data base,	*/
/*				initialize it's data vector in one of several	*/
/*				ways:											*/
/*																*/
/*	input:		db 		- pointer to database					*/
/*																*/
/*	output:		none											*/
/****************************************************************/
int simulate_data( DB *db, int dims, int count, int clust)
{
	long	i, j;
	int num_clusters;
	double	step;
	MATRIX	*A;

	DEBUG( "simulate_data" );
	
    if( !db )
        return( ERROR );

	if( db->dat.data )
		free_data( db );

	switch( db->dat.data_init_mode )
	{
		case SIM_TRI:
			db->dat.num_points = 3;
			db->dat.data_dims = 2;
			db->draw_mode = LABELS | ATTRACTIONS;
//			set_mode_toggles( db->draw_mode );

			db->name = mstrdup( "Triangle" );

			/*
			 * allocate data for database
			 */
			if( alloc_data( db ) == ERROR )
				return( ERROR );

			A = db->dat.A;

			/*
			 * set up triangle test case
			 */
			db->dat.data[0].label = mstrdup( "A" );
			A->data[0][0] = 0;
			A->data[0][1] = 0;
			
			db->dat.data[1].label = mstrdup( "B" );
			A->data[1][0] = 1;
			A->data[1][1] = 0;
			
			db->dat.data[2].label = mstrdup( "C" );
			A->data[2][0] = 0.5;
			A->data[2][1] = 0.875;
			break;
		
		case SIM_CUBE:
			db->dat.num_points = 8;
			db->dat.data_dims = 3;
			db->animate.max_iteration = 10;
			db->draw_mode = LABELS | LINKS;
//			set_mode_toggles( db->draw_mode );

			db->name = mstrdup( "Cube" );

			/*
			 * allocate data for database
			 */
			if( alloc_data( db ) == ERROR )
				return( ERROR );

			A = db->dat.A;

			/*
			 * set up cube test case
			 */
			 
			/* base */
			db->dat.data[0].label = mstrdup( "0" );
			A->data[0][0] = 0;
			A->data[0][1] = 0;
			A->data[0][2] = 0;
			db->dat.data[0].connections[0] = 1;
			db->dat.data[0].connections[1] = 4;
			db->dat.data[0].num_connections = 2;
			
			db->dat.data[1].label = mstrdup( "1" );
			A->data[1][0] = 1;
			A->data[1][1] = 0;
			A->data[1][2] = 0;
			db->dat.data[1].connections[0] = 2;
			db->dat.data[1].connections[1] = 5;
			db->dat.data[1].num_connections = 2;
			
			db->dat.data[2].label = mstrdup( "2" );
			A->data[2][0] = 1;
			A->data[2][1] = 0;
			A->data[2][2] = 1;
			db->dat.data[2].connections[0] = 3;
			db->dat.data[2].connections[1] = 6;
			db->dat.data[2].num_connections = 2;
			
			db->dat.data[3].label = mstrdup( "3" );
			A->data[3][0] = 0;
			A->data[3][1] = 0;
			A->data[3][2] = 1;
			db->dat.data[3].connections[0] = 0;
			db->dat.data[3].connections[1] = 7;
			db->dat.data[3].num_connections = 2;
			
			/* top */
			db->dat.data[4].label = mstrdup( "4" );
			A->data[4][0] = 0;
			A->data[4][1] = 1;
			A->data[4][2] = 0;
			db->dat.data[4].connections[0] = 5;
			db->dat.data[4].num_connections = 1;
			
			db->dat.data[5].label = mstrdup( "5" );
			A->data[5][0] = 1;
			A->data[5][1] = 1;
			A->data[5][2] = 0;
			db->dat.data[5].connections[0] = 6;
			db->dat.data[5].num_connections = 1;
			
			db->dat.data[6].label = mstrdup( "6" );
			A->data[6][0] = 1;
			A->data[6][1] = 1;
			A->data[6][2] = 1;
			db->dat.data[6].connections[0] = 7;
			db->dat.data[6].num_connections = 1;
			
			db->dat.data[7].label = mstrdup( "7" );
			A->data[7][0] = 0;
			A->data[7][1] = 1;
			A->data[7][2] = 1;
			db->dat.data[7].connections[0] = 4;
			db->dat.data[7].num_connections = 1;
			break;

		case SIM_TETRAHEDRON:
			db->dat.num_points = 4;
			db->dat.data_dims = 3;
			db->name = mstrdup( "Tetrahedron" );

			/*
			 * allocate data for database
			 */
			if( alloc_data( db ) == ERROR )
				return( ERROR );

			A = db->dat.A;

			/* base */
			db->dat.data[0].label = mstrdup( "0" );
			A->data[0][0] = 0;
			A->data[0][1] = 0;
			A->data[0][2] = 0;
			db->dat.data[0].connections[0] = 1;
			db->dat.data[0].connections[1] = 2;
			db->dat.data[0].connections[2] = 3;
			db->dat.data[0].num_connections = 3;
			
			db->dat.data[1].label = mstrdup( "1" );
			A->data[1][0] = 0;
			A->data[1][1] = 1;
			A->data[1][2] = 0;
			db->dat.data[1].connections[0] = 0;
			db->dat.data[1].connections[1] = 2;
			db->dat.data[1].connections[2] = 3;
			db->dat.data[1].num_connections = 3;
			
			db->dat.data[2].label = mstrdup( "2" );
			A->data[2][0] = sqrt(3.0)/2.0;
			A->data[2][1] = 0.5;
			A->data[2][2] = 0;
			db->dat.data[2].connections[0] = 0;
			db->dat.data[2].connections[1] = 1;
			db->dat.data[2].connections[2] = 3;
			db->dat.data[2].num_connections = 3;
			
			db->dat.data[3].label = mstrdup( "3" );
			A->data[3][0] = sqrt(3.0)/6.0;
			A->data[3][1] = 0.5;
			A->data[3][2] = sqrt(2.0/3.0);
			db->dat.data[3].connections[0] = 0;
			db->dat.data[3].connections[1] = 1;
			db->dat.data[3].connections[2] = 2;
			db->dat.data[3].num_connections = 3;
			
			break;
				
		case SIM_HYPERCUBE:
			break;
			
		case SIM_CLUSTER:
		default:
                        db->dat.num_points = count;
                        db->dat.data_dims = dims;
                        db->dat.num_clusters = clust;

			db->name = mstrdup( "Random Cluster" );

			/*
			 * allocate data for database
			 */
			if( alloc_data( db ) == ERROR )
				return( ERROR );

			A = db->dat.A;

			num_clusters = db->dat.num_clusters;
			step = 1.0/num_clusters;
				
			/*
			 * init database with random values
			 */
			for( i = 0; i < db->dat.num_points ; i++ )
            {
                for( j = 0; j < db->dat.data_dims; j++ )
				{
					if( nrand() < db->dat.clust_prob )
					   A->data[i][j] = (nrand()/num_clusters)+(i%num_clusters)
																		* step;
					else
					   A->data[i][j] = nrand();
				}
			}
			break;
	}
	
	/*
	 * save unnormalized data before we normalize
	 */
	copy_matrix( db->dat.R, db->dat.A );
	normalize_matrix( db->dat.A );

	/*
	 * now that we have data - calculate
	 * N-dimensional distances - ONE TIME ONLY :-)
	 */
	calc_all_distances( db );

	/*
     * calculate screen coords after
     * clusters have been computed
	 * and draw data to show user
     */
    setup_screen_coords( db );

	return( SUCCESS );
}

/****************************************************************/
/*	function:	report_db() - print out stats on database to	*/
/*				ASCII text file									*/
/*																*/
/*	input:		db		- pointer to database					*/
/*																*/
/*	output:		none											*/	
/****************************************************************/
void report_db( void *data )
{
	DB		*db = (DB *)data;
	long	i, num;
	long	count, total;
	long	min, max;
	FILE	*fp;

	DEBUG( "report_db" );

	if( !db || !db->files.reportfp )
		return;
		
	fp = db->files.reportfp;

	/*
	 * print database fields
	 */
	fprintf( fp, "\nGeneral DB info:\n\n" );
	fprintf( fp, "Name: %s\n", db->name );
	fprintf( fp, "Input file: %s\n", db->files.infile );
	fprintf( fp, "Output file: %s\n", db->files.outfile );
	fprintf( fp, "Num Points: %ld\n", db->dat.num_points );
	fprintf( fp, "Num Clusters: %ld\n", db->dat.num_clusters );
	fprintf( fp, "Num data dimensions: %ld\n", db->dat.data_dims );
	fprintf( fp, "Num display dimensions: %ld\n", db->dat.display_dims );
	fprintf( fp, "Max Iteration: %ld\n", db->animate.max_iteration );

	fprintf( fp, "\nDB modes:\n\n" );
	fprintf( fp, "Max Children: %ld\n", db->dat.max_children );
	fprintf( fp, "Data init mode: %d\n", db->dat.data_init_mode );
	fprintf( fp, "Coord init mode: %d\n", db->mds.display_init_mode );
	fprintf( fp, "Attractor select mode: %d\n", db->mds.select_mode );
	fprintf( fp, "Output mode: %d\n", db->output_mode );
	fprintf( fp, "Draw mode: %d\n", db->draw_mode );
	fprintf( fp, "Plot mode: %d\n", db->plot_mode );

	fprintf( fp, "\nDB state variables:\n\n" );
	fprintf( fp, "Distance function: %d\n", db->mds.dist_func );
	fprintf( fp, "Num bonds: %d\n", db->mds.num_bonds );
	fprintf( fp, "Cluster Probability: %f\n", db->dat.clust_prob );

	fprintf( fp, "\nDB stats:\n\n" );
	fprintf( fp, "Num nD distances: %ld\n", db->stats.num_ndist );
	fprintf( fp, "Min nD distance: %f\n", db->stats.min_ndist );
	fprintf( fp, "Max nD distance: %f\n", db->stats.max_ndist );
	fprintf( fp, "Mean nD distances: %f\n", db->stats.mean_ndist );

	/*
	 * collect some stats on DB
	 */
	total = 0;
	count = 0;
	min = db->dat.num_points;
	max = 0;

	for( i = 0; i < db->dat.A->height ; i++ )
	{
		num += db->dat.data[i].num_bonds;
		total += num;
		
		min = MIN( min, num );
		max = MAX( max, num );

		count++;
	}
	
	fprintf( fp, "(counting)Num nD distances: %ld\n", total );
	fprintf( fp, "mean nD distances per datum: %f\n", total/(double)count );

}

/************************************************************************/
/*	function:	quit() - frees program memory and exits					*/
/*																		*/
/*	input:																*/
/*		Widget		- integer id of widget that initiated callback		*/
/*		data		- void pointer to database							*/
/*																		*/
/*	output:		none													*/	
/************************************************************************/
void quit(void *data)
{
	DB	*db = (DB *)data;

	DEBUG( "quit" );

	/*
     * clean up
     */
    free_db( db );

	exit(0);
}
