#include <math.h>
#include <sstream>
#include <ctime>
#include "DBColumnIterator.h"


DBColumnIterator::DBColumnIterator(DBManager* dbManager, std::string dataTableName, 
             DimensionSpace* allDimSpace, DimensionSpace* loadDimSpace, int rowNum, Query* query)

{
    this->dbManager = dbManager;
    this->dataTableName = dataTableName;
    this->allDimSpace = allDimSpace;
    this->loadDimSpace = loadDimSpace;
    this->rowNum = rowNum;
    this->query = query;

    queryMode = query->GetQueryMode();
    groupingMode = query->GetGroupingMode();
    isMaskable = query->IsMaskable();
    isIncreasingOrder = query->IsIncreasingOrder();
    sampling = query->IsSampling();
    approximateSize = query->GetApproximateSampleSize();
    samplingMode = query->GetSamplingMode();

    loadDimNum = loadDimSpace->GetSize();
    if (loadDimNum == 0)
    {
        noquery = true;
        if (groupingMode != MASK_GROUP)
            isMaskable = false;
    }
    else
        noquery = false;

    if (groupingMode == MASK_GROUP)
        noquery = false;

    if (noquery)
    {
        buffer = NULL;
        loadDimSpace = NULL;
        docColumn = new DocColumn();
        count = 0;
        return;
    }

    if (rowNum <= approximateSize)
        this->sampling = false;

    docColumn = new DocColumn();

    if (queryMode == SCATTER_QUERY)
        docColumn->SetMode(DUAL);
    else if (queryMode == XY_QUERY)
        docColumn->SetMode(TRIPLE);

    if (isMaskable)
        docColumn->SetMaskMode(true);
	    
    if (queryMode == XY_QUERY)
    {
        Dimension* dim = allDimSpace->GetDimensionByID(query->GetXDimensionID());
        loadDimSpace->AddDimension(dim);
        dim = allDimSpace->GetDimensionByID(query->GetYDimensionID());
        loadDimSpace->AddDimension(dim);
    }
    else if (queryMode == SCATTER_QUERY)
    {
        Dimension* dim = allDimSpace->GetDimensionByID(query->GetScatterDimensionID());
        loadDimSpace->AddDimension(dim);
    }

//    if (groupingMode == MASK_GROUP)
    if (isMaskable)
    {
        Dimension* dim = allDimSpace->GetDimensionByID(query->GetGroupByDimensionID());
        loadDimSpace->AddDimension(dim);
    }

    loadDimNum = loadDimSpace->GetSize();

    if (loadDimNum != 0)
    {
        if (groupingMode != NONE_GROUP && groupingMode != MASK_GROUP)
            columnSize = GetColumnSizeInGroupMode();
        else
            columnSize = GetColumnSizeInNormalMode();

        docColumn->SetSize(columnSize);

        buffer = new double* [loadDimNum];
        for (int i = 0; i < loadDimNum; i++)
        {
            buffer[i] = new double [columnSize];
        }
        if (groupingMode != NONE_GROUP && groupingMode != MASK_GROUP)
            LoadColumnsInGroupMode();
        else 
            LoadColumnsInNormalMode();

        if (queryMode == XY_QUERY)
        {
            for (i = 0; i < columnSize; i ++)
            {
                docColumn->SetXData(i, buffer[loadDimNum-2][i]); 
                docColumn->SetYData(i, buffer[loadDimNum-1][i]); 
            }
            delete [] buffer[loadDimNum-2];
            buffer[loadDimNum-2] = NULL;
            delete [] buffer[loadDimNum-1];
            buffer[loadDimNum-1] = NULL;
        }
        else if (queryMode == SCATTER_QUERY)
        {
            for (i = 0; i < columnSize; i ++)
            {
                docColumn->SetBaseData(i, buffer[loadDimNum-1][i]); 
            }
            delete [] buffer[loadDimNum-1];
            buffer[loadDimNum-1] = NULL;
        }

        if (isMaskable)
        {
            Dimension* dim = allDimSpace->GetDimensionByID(query->GetGroupByDimensionID());
            double value1 = dim->GetNorValue(query->GetGroupValue());
            for (i = 0; i < columnSize; i ++)
            {
                if (fabs(buffer[loadDimNum-1][i]-value1) < 1e-5)
                    docColumn->SetMaskData(i, 0);
                else
                    docColumn->SetMaskData(i, 1);
            }
            delete [] buffer[loadDimNum-1];
            buffer[loadDimNum-1] = NULL;
        }
    }
    else
    {
        buffer = NULL;
    }

    count = 0;
}	


DBColumnIterator::~DBColumnIterator()
{
    if (buffer != NULL)
    {
        for (int i = 0; i < loadDimNum; i++)
        {
            if (buffer[i] != NULL)
                delete [] buffer[i];
        }
        delete [] buffer;
    }

    delete docColumn;
}

bool DBColumnIterator::LoadColumnsInNormalMode()
{
    int i, j;

    int orderByDimID = query->GetOrderByDimensionID();

    Dimension** dims;
    dims = new Dimension *[loadDimNum];
    for (i = 0; i < loadDimNum; i++)
    {
        dims[i] = loadDimSpace->GetDimension(i);
    }

    int *orders = new int [loadDimNum];

    IIHash hash;
    int id;
    //need to remove redundent dimensions for query
    DimensionSpace* tempSpace = new DimensionSpace();

    int count = 0;
    for (i = 0; i < loadDimNum; i ++)
    {
        id = loadDimSpace->GetDimension(i)->GetDimensionID();
        if (hash.find(id) == hash.end())
        {
            hash[id] = count;
            tempSpace->AddDimension(loadDimSpace->GetDimension(i));
            orders[i] = count;
            count ++;
        }
        else 
            orders[i] = hash[id];
    }
    int queryDimNum = count;

    std::ostringstream oss;
    oss << "select ";
    for (i = 0; i < tempSpace->GetSize(); i ++)
    {
        oss << "t" << tempSpace->GetDimension(i)->GetDimensionID() << ", ";
    }

    delete tempSpace;

    oss << "itemID from " << dataTableName;

    if (sampling)
        oss << " ta " << "where ta.randomColumn > " << query->GetStart() << " and ta.randomColumn < " << query->GetEnd();
    
    if ( orderByDimID >= 0 )
        oss << " order by t" << orderByDimID;
    else
        oss << " order by itemID";
    if (!isIncreasingOrder)
        oss << " DESC";
    
    ResultSet* re = dbManager->ExecuteQuery(oss.str());

    for (i = 0; i < columnSize; i++)
	{
		re->next();
	    for (j = 0; j < loadDimNum; j++)
	    {
           if (dims[j]->GetType() == DOUBLE_DIMENSION || dims[j]->GetType() == BINARY_DIMENSION )
           {
               buffer[j][i] = dims[j]->GetNorValue(re->getDouble(orders[j]));
           }
           else
           {
               buffer[j][i] = dims[j]->GetNorValue(re->getString(orders[j]));
           }
        }
        docColumn->SetItemID(i, re->getInt(queryDimNum));
    }
    delete re;
    delete dims;
    delete orders;

    return true;
}


int DBColumnIterator::GetColumnSizeInNormalMode()
{
    if (sampling == false)
        return rowNum;
    else if (samplingMode == RANDOM_SAMPLING)
        return GetRandomSamplingColumnSize();
    else //if (samplingMode == BIASED_SAMPLING)
        return GetBiasedSamplingColumnSize();
}

int DBColumnIterator::GetRandomSamplingColumnSize()
{
    int size;
    if (rowNum > approximateSize)
    {
        std::ostringstream oss;
        oss << "select count(*)" << " from " << dataTableName << " ta "
            << "where ta.randomColumn > " << query->GetStart() << " and ta.randomColumn < " << query->GetEnd();
        ResultSet* re = dbManager->ExecuteQuery(oss.str());
	    re->next();
//for mysql
//        size = re->getInt(0);
//for oracle
        size = re->getDouble(0);
        delete re;
    }
    else
        size = rowNum;

    return size;
}

int DBColumnIterator::GetBiasedSamplingColumnSize()
{
    int size = 0;
/*    
    std::string* values;
    std::string* dataTableNames;
    int groupNum;
    double* rates;
    int temp;
    for (int i = 0; i < groupNum; i++)
    {
//        srand( (unsigned)time( NULL ));
    
        gStart[i] = (double)rand()/RAND_MAX * (1-rates[i]);
        gEnd[i] = gStart[i] + rates[i];

        std::ostringstream oss;
        oss << "select count(*)" << " from " << dataTableNames[i] << " ta "
            << "where t" << sorDim << "=" << values[i].c_str() << " and " 
            "ta.randomColumn > " << start << " and ta.randomColumn < " << end;
        ResultSet* re = dbManager->ExecuteQuery(oss.str());
	    re->next();
        temp = re->getInt(0);
        size += temp;
        delete re;
    }*/
    return size;
}


void* DBColumnIterator::GetNext()
{
    if (count == loadDimNum)
        return NULL;

    int dimID = loadDimSpace->GetDimension(count)->GetDimensionID();
	docColumn->SetDimensionID(dimID);

    for (int i = 0; i < columnSize; i ++)
    {
        docColumn->SetData(i, buffer[count][i]); 
    }

    delete [] buffer[count];
    buffer[count] = NULL;

    count ++;
	return docColumn;
}


bool DBColumnIterator::IsDone()
{
	if (count >= loadDimNum)
		return true;
	else
		return false;
}

int DBColumnIterator::GetColumnSizeInGroupMode()
{
    std::ostringstream oss;
    std::string groupValue = query->GetGroupValue();
    int groupDimID = query->GetGroupByDimensionID();

    oss << "select count(*)" << " from " << dataTableName
        << " where t" << groupDimID << " = \"" << groupValue.c_str() << "\"";
    
    ResultSet* re = dbManager->ExecuteQuery(oss.str());
	re->next();
//for mysql
//    int size = re->getInt(0);
    int size = re->getDouble(0);

    return size;
}

bool DBColumnIterator::LoadColumnsInGroupMode()
{
    int i, j;

    std::string groupValue = query->GetGroupValue();
    int groupDimID = query->GetGroupByDimensionID();
    int orderbyDimID = query->GetOrderByDimensionID();

    Dimension** dims;
    dims = new Dimension *[loadDimNum];
    for (i = 0; i < loadDimNum; i++)
    {
        dims[i] = loadDimSpace->GetDimension(i);
    }

    std::ostringstream oss;
    oss << "select ";
    for (i = 0; i < loadDimNum-1; i ++)
    {
        oss << "t" << loadDimSpace->GetDimension(i)->GetDimensionID() << ", ";
    }
    oss << "t" << loadDimSpace->GetDimension(loadDimNum-1)->GetDimensionID() 
        << " from " << dataTableName << " where t" << groupDimID << " = \"" << groupValue.c_str() << "\""; 

    if ( orderbyDimID >= 0 )
        oss << " order by t" << orderbyDimID;
    else
        oss << " order by itemID";
    if (!isIncreasingOrder)
        oss << " DESC";
    
    ResultSet* re = dbManager->ExecuteQuery(oss.str());

    for (i = 0; i < columnSize; i++)
	{
		re->next();
	    for (j = 0; j < loadDimNum; j++)
	    {
           if (dims[j]->GetType() == DOUBLE_DIMENSION || dims[j]->GetType() == BINARY_DIMENSION )
           {
               buffer[j][i] = dims[j]->GetNorValue(re->getDouble(j));
           }
           else
           {
               buffer[j][i] = dims[j]->GetNorValue(re->getString(j));
           }
        }
    }
    delete re;
    delete dims;

    return true;
}
