/****************************************************************/
/* headers                                                      */
/****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <limits.h>
#include "general.h"
#include "util.h"
#include "nvect.h"

/****************************************************************/
/* defines                                                      */
/****************************************************************/
	
/****************************************************************/
/* typedefs                                                     */
/****************************************************************/

/****************************************************************/
/* external functions                                           */
/****************************************************************/
		
/****************************************************************/
/* function prototypes                                          */
/****************************************************************/
/*
 * utility functions
 */
double *alloc_nvect( long veclen );
int init_display_mask( long veclen);
void init_nvect( double vect[], long veclen );
void copy_nvect( double vect1[], double vect2[], long veclen );
void set_distance_func( int mode );

/*
 * distance functions
 */
double euclid_dist( double vect1[], double vect2[], int mask[], long veclen );
double manhatten_dist( double vect1[], double vect2[], int mask[], long veclen);
double minkowski_dist( double vect1[], double vect2[], int mask[], long veclen);
double cosine_dist( double vect1[], double vect2[],int mask[], long veclen );
double correlate_nvect( double vect1[], double vect2[], int mask[],long veclen);

/*
 * n-vector math functions
 */
double vec_len( double vect[], int mask[], long veclen );
void norm_nvect( double vect[], int mask[], long veclen, double result[] );
void div_nvect( double num, double vect[], long veclen, double result[] );
void mult_nvect( double num, double vect[], long veclen, double result[] );
void add_nvect( double vect1[], double vect2[], long veclen, double result[] );
void sub_nvect( double vect1[], double vect2[], long veclen, double result[] );
double min_nvect( double vect[], int mask[], long veclen );
double max_nvect( double vect[], int mask[], long veclen );
long mini_nvect( double vect[], int mask[], long veclen );
long maxi_nvect( double vect[], int mask[], long veclen );
double mean_nvect( double vect[], int mask[], long veclen );
double median_nvect( double vect[], int mask[], long veclen );
void sort_nvect( double vect[], long veclen );
void rand_nvect( double vect[], long veclen );

/****************************************************************/
/* global variables												*/
/****************************************************************/
double (*distance)( double [], double [], int [], long ) = euclid_dist;
int *display_mask;
double r;

/****************************************************************/
/*	function	alloc_nvect() -allocate memory for n-vector		*/
/*																*/
/*	input:		veclen	- length of vector to create			*/				
/*																*/
/*	output:		output vector of doubles						*/
/****************************************************************/
double *alloc_nvect( long veclen )
{
	return( (double *)mmalloc(veclen * sizeof(double)) );
}

/****************************************************************/
/*	function	init_display_mask() - initializes the mask that */
/*              determines whether a given display dimension is */
/*              included in the distance calculations that are  */
/*              the heart of the MDS procedure. The default     */
/*              state is for all the dimensions to be "enabled" */
/*              i.e. to have '1's in every cell of the mask     */
/*              NOTE: this mask is actually never changed       */
/*              it is just filled with '1's and then ignored -  */
/*              it is just needed because the distance          */
/*              functions require a mask to their interfaces    */
/*              actually a                                      */
/*																*/
/*	input:		in_mask	- input vector of 1's and 0's			*/
/*				veclen	- length of mask vector					*/
/*																*/
/*	output:		none											*/
/****************************************************************/
int init_display_mask( long veclen )
{
	long	i;

	if( (display_mask = (int *)mmalloc(veclen*sizeof(int))) == NULL )
		return( ERROR );

	for( i = 0; i < veclen; i++ )
		display_mask[i] = 1;

	return( SUCCESS );
}

/****************************************************************/
/*	function	init_nvect() - initialize n-vector to all zeros	*/
/*																*/
/*	input:		vect	- input vector of doubles to initialize	*/
/*				veclen	- length of input vector				*/				
/*																*/
/*	output:		none											*/
/****************************************************************/
void init_nvect( double vect[], long veclen )
{
	long	i;
		
	for( i = 0; i < veclen; i++ )
		vect[i] = 0.0;
}

/****************************************************************/
/*	function:	copy_nvect() - copy n-vector to another			*/
/*																*/
/*	input:		vect1	- n-vector to copy to					*/
/*				vect2	- n-vector to copy from					*/
/*				veclen	- length of vector						*/
/*																*/
/*	output:		none											*/
/****************************************************************/
void copy_nvect( double vect1[], double vect2[], long veclen )
{
	long	i;
	
	for( i = 0; i < veclen; i++ )
		vect1[i] = vect2[i];	 
}

/****************************************************************/
/*  function    set_distance_func() - set function used to		*/
/*				compute distance or similarity					*/
/*                                                              */
/*  input:      mode	- integer id of distance function		*/
/*  output:     none                                            */
/****************************************************************/
void set_distance_func( int mode )
{
    switch( mode )
    {
		case COSINE:
				distance = cosine_dist;
				break;

		case CORRELATE:
				distance = correlate_nvect;
				break;

        case MINKOWSKI:
                distance = minkowski_dist;
                break;

        case MANHATTAN:
                distance = manhatten_dist;
                break;

        case EUCLIDIAN:
        default:
                distance = euclid_dist;
                break;
    }
}
/****************************************************************/
/*	function	euclid_dist() - calculate distance between two	*/
/*				n-dimensional vectors							*/
/*																*/
/*	input:		vect1	- input vector of doubles				*/
/*				vect2	- second input vector of doubles		*/
/*				mask	- array of integer flags enable/disable	*/
/*				veclen	- length of input vector				*/				
/*																*/
/*	output:		length of vector as double						*/
/****************************************************************/
double euclid_dist( double vect1[], double vect2[], int mask[], long veclen )
{
	double 	len = 0.0;
	double	delta, result;
	long	i;
	
	for( i = 0; i < veclen; i++ )
	{
		if( mask[i] )
		{
			delta = vect1[i] - vect2[i];
			len += SQR(delta);
		}
	}
	
	result = sqrt(len);
	
	return( result );
}

/****************************************************************/
/*	function	manhatten_dist() - calculate distance between 2	*/
/*				n-dimensional vectors							*/
/*																*/
/*	input:		vect1	- input vector of doubles				*/
/*				vect2	- second input vector of doubles		*/
/*				mask	- array of integer flags enable/disable	*/
/*				veclen	- length of input vector				*/				
/*																*/
/*	output:		length of vector as double						*/
/****************************************************************/
double manhatten_dist( double vect1[], double vect2[], int mask[], long veclen )
{
	double 	sum = 0.0;
	long	i;
	
	for( i = 0; i < veclen; i++ )
	{
		if( mask[i] )
			sum += ABS(vect1[i]-vect2[i]);
	}
	
	return( sum );
}

/****************************************************************/
/*	function	minkowski_dist() - calculate distance between 	*/
/*				two n-dimensional vectors						*/
/*																*/
/*	input:		vect1	- input vector of doubles				*/
/*				vect2	- second input vector of doubles		*/
/*				mask	- array of integer flags enable/disable	*/
/*				veclen	- length of input vector				*/				
/*																*/
/*	output:		length of vector as double						*/
/****************************************************************/
double minkowski_dist( double vect1[], double vect2[], int mask[], long veclen )
{
	double 	len = 0.0;
	long	i;
	
	for( i = 0; i < veclen; i++ )
	{
		if( mask[i] )
			len += pow( vect1[i] - vect2[i], r );
	}
	
	return(  pow( len, 1.0/r ) );
}

/****************************************************************/
/*  function    cosine_dist() - calculate cosine between 2 		*/
/*              n-dimensional vectors                           */
/*                                                              */
/*  input:      vect1   - input vector of doubles               */
/*              vect2   - second input vector of doubles        */
/*              mask    - array of ints for enable/disable      */
/*              veclen  - length of input vector                */
/*                                                              */
/*  output:     length of vector as double                      */
/****************************************************************/
double cosine_dist( double vect1[], double vect2[], int mask[], long veclen )
{
    double  sum = 0.0;
	double	len1, len2;
    long    i;

	/*
	 * calculate dot product of vectors
	 */
    for( i = 0; i < veclen; i++ )
    {
        if( mask[i] )
            sum += (vect1[i] * vect2[i]);
    }

	/*
	 * calculate vector lengths to divide dot product
	 */
	len1 = vec_len( vect1, mask, veclen );
	if( FLT_ZERO(len1) )
		len1 = 0.0001;

	len2 = vec_len( vect2, mask, veclen );
	if( FLT_ZERO(len2) )
		len2 = 0.0001;

    return( sum/(len1*len2) );
}

/****************************************************************/
/*  function:   correlate_nvect() - calculates the         		*/
/*              correlation between two n-dimensional vectors	*/
/*              Uses formula for correlation, "Pearson's r" 	*/
/*				from  Numerical Recipies (14.5.1) (p. 636):     */
/*                                                              */
/*                       --                 --                  */
/*                       \ (x - X_mean) *   \ (y - Y_mean)      */
/*                       /                  /                   */
/*                       --                 --                  */
/*                      x in X              y in Y              */
/*              r = ---------------------------------------     */
/*                 ------------------      -----------------    */
/*                 | --                    | --                 */
/*              -  | \ (x - X_mean)^2 * -  | \ (y - Y_mean)^2   */
/*               \ | /                   \ | /                  */
/*                \| --                   \| --                 */
/*                   x in X                  y in Y             */
/*                                                              */
/*  input:      vect1   - input vector of doubles               */
/*              vect2   - second input vector of doubles        */
/*              mask    - array of ints for enable/disable      */
/*              veclen  - length of input vector                */
/*                                                              */
/*  output:     double correlation between two n-vectors		*/
/****************************************************************/
double correlate_nvect( double vect1[], double vect2[], int mask[], long veclen)
{
	double  mean1, mean2;
    double  delta1, delta2;
    double  sum1 = 0;
    double  sum2 = 0;
    double  numer = 0;
    double  denom = 0;
    long    i;

    /*
     * calculate mean of two n-vectors
     */
    mean1 = mean_nvect( vect1, mask, veclen );
    mean2 = mean_nvect( vect2, mask, veclen );

    /*
     * sum values for each vector
     */
    for( i = 0; i < veclen; i++ )
    {
		if( mask[i] )
		{
			delta1 = (vect1[i] - mean1);
			delta2 = (vect2[i] - mean2);

			numer += (delta1 * delta2);

			sum1 += SQR(delta1);
			sum2 += SQR(delta2);
		}
	}

	denom = sqrt(sum1) * sqrt(sum2);

    return( (FLT_ZERO(denom)) ? 0.0 : numer/denom );
}

/****************************************************************/
/*	function	vec_len() - calculate length of n-dimensional	*/
/*              vector											*/
/*																*/
/*	input:		vect	- input vector of doubles				*/
/*              mask    - array of ints for enable/disable      */
/*				veclen	- length of input vector				*/				
/*																*/
/*	output:		length of vector as double						*/
/****************************************************************/
double vec_len( double vect[], int mask[], long veclen )
{
	double 	len = 0.0;
	long	i;
	
	for( i = 0; i < veclen; i++ )
	{
		if( mask[i] )
			len += SQR(vect[i]);
	}
	
	return( sqrt(len) );
}

/****************************************************************/
/*	function	norm_nvect() - normalize n-dimensional vector	*/
/*																*/
/*	input:		vect	- input vector of doubles				*/
/*              mask    - array of ints for enable/disable      */
/*				veclen	- length of input vector				*/
/*				result	- output vector of doubles				*/	
/*																*/
/*	output:		none											*/
/****************************************************************/
void norm_nvect( double vect[], int mask[], long veclen, double result[] )
{
	double 	len = vec_len( vect, mask, veclen );
	double	invlen;
	long	i;
	
	if( FLT_ZERO(len) )
		len = 0.00001;

	invlen = 1.0/len;
	
	for( i = 0; i < veclen; i++ )
		result[i] = vect[i] * invlen;

}

/****************************************************************/
/*	function:	div_nvect() - divides vector by a scalar value	*/
/*																*/
/*	input:		num		- scalar to divide by					*/
/*				vect	- n-vector to divide					*/
/*				veclen	- length of vector						*/
/*				result	- vector to hold resulting doubles		*/
/*																*/
/*	output:		none											*/
/****************************************************************/
void div_nvect( double num, double vect[], long veclen, double result[] )
{
	long	i;
	double	inv_num;
	
	if( FLT_ZERO( num ) )
		return;
		
	inv_num = 1.0/num;
	
	for( i = 0; i < veclen; i++ )
		result[i] = inv_num * vect[i];	 
}

/****************************************************************/
/*	function:	mult_nvect() - multipies vector by a scalar val */
/*																*/
/*	input:		num		- scalar to multiply by					*/
/*				vect	- n-vector to multiply					*/
/*				veclen	- length of vector						*/
/*				result	- vector to hold resulting doubles		*/
/*																*/
/*	output:		none											*/
/****************************************************************/
void mult_nvect( double num, double vect[], long veclen, double result[] )
{
	long	i;
	
	for( i = 0; i < veclen; i++ )
		result[i] = num * vect[i];	 
}

/****************************************************************/
/*	function:	add_nvect() - adds two n-vectors				*/
/*																*/
/*	input:		vect1	- n-vector to add						*/
/*				vect2	- other n-vector to add					*/
/*				veclen	- length of vector						*/
/*				result	- vector to hold resulting doubles		*/
/*																*/
/*	output:		none											*/
/****************************************************************/
void add_nvect( double vect1[], double vect2[], long veclen, double result[] )
{
	long	i;
	
	for( i = 0; i < veclen; i++ )
		result[i] = vect1[i] + vect2[i];	 
}

/****************************************************************/
/*	function:	sub_nvect() - subtract two n-vectors			*/
/*																*/
/*	input:		vect1	- n-vector to subtract from				*/
/*				vect2	- n-vector to subtract					*/
/*				veclen	- length of vector						*/
/*				result	- vector to hold resulting doubles		*/
/*																*/
/*	output:		none											*/
/****************************************************************/
void sub_nvect( double vect1[], double vect2[], long veclen, double result[] )
{
	long	i;
	
	for( i = 0; i < veclen; i++ )
		result[i] = vect1[i] - vect2[i];	 
}

/****************************************************************/
/*	function:	min_nvect() - find minimum value in an n-vector	*/
/*																*/
/*	input:		vect	- n-vector to search					*/
/*              mask    - array of ints for enable/disable      */
/*				veclen	- length of vector						*/
/*																*/
/*	output:		double minimum scalar component of vector		*/
/****************************************************************/
double min_nvect( double vect[], int mask[], long veclen )
{
	long	i = mini_nvect(vect, mask, veclen);
	
	return( vect[i] );
}

/****************************************************************/
/*	function:	max_nvect() - find maximum value in an n-vector	*/
/*																*/
/*	input:		vect	- n-vector to search					*/
/*              mask    - array of ints for enable/disable      */
/*				veclen	- length of vector						*/
/*																*/
/*	output:		double maximum scalar component of vector		*/
/****************************************************************/
double max_nvect( double vect[], int mask[], long veclen )
{
	long	i = maxi_nvect(vect, mask, veclen);
	
	return( vect[i] );
}

/****************************************************************/
/*	function:	mini_nvect() - find index of min value in an 	*/
/*							  n-vector							*/
/*																*/
/*	input:		vect	- n-vector to search					*/
/*              mask    - array of ints for enable/disable      */
/*				veclen	- length of vector						*/
/*																*/
/*	output:		integer index of minimum scalar of vector		*/
/****************************************************************/
long mini_nvect( double vect[], int mask[], long veclen )
{
	long	i, mini = 0;
	double 	min = DBL_MAX;
	
	/*
	 * find first enabled cell
	 */
	for( i = 0; i < veclen; i++ )
	{
		if( mask[i] )
		{
			min = vect[i];
			mini = i;
			break;
		}
	}

	/*
	 * look through rest of cells for min
	 */
	for( ++i; i < veclen; i++ )
	{
		if( mask[i] && vect[i] < min )
		{
			min = vect[i];
			mini = i;
		}
	}
	
	return( mini );
}

/****************************************************************/
/*	function:	maxi_nvect() - find index of max value in an 	*/
/*							  n-vector							*/
/*																*/
/*	input:		vect	- n-vector to search					*/
/*              mask    - array of ints for enable/disable      */
/*				veclen	- length of vector						*/
/*																*/
/*	output:		integer index of maximum scalar of vector		*/
/****************************************************************/
long maxi_nvect( double vect[], int mask[], long veclen )
{
	long	i, maxi = 0;
	double 	max = DBL_MIN;
	
	/*
	 * find first enabled cell
	 */
	for( i = 0; i < veclen; i++ )
	{
		if( mask[i] )
		{
			max = vect[i];
			maxi = i;
			break;
		}
	}

	/*
	 * look through rest of cells for max
	 */
	for( ++i; i < veclen; i++ )
	{
		if( mask[i] && vect[i] > max )
		{
			max = vect[i];
			maxi = i;
		}
	}
	
	return( maxi );
}

/****************************************************************/
/*	function:	mean_nvect() - find mean value in an n-vector	*/
/*																*/
/*	input:		vect	- n-vector to process					*/
/*              mask    - array of ints for enable/disable      */
/*				veclen	- length of vector						*/
/*																*/
/*	output:		double mean of vector							*/
/****************************************************************/
double mean_nvect( double vect[], int mask[], long veclen )
{
	long	i, count = 0;
	double	sum = 0.0;

	for( i = 0; i < veclen; i++ )
	{
		if( mask[i] )
		{
			sum += vect[i];
			count++;
		}
	}

	return( sum/count );
}

/****************************************************************/
/*	function:	median_nvect() - find median value in an 		*/
/*								n-vector						*/
/*																*/
/*	input:		vect	- n-vector to search					*/
/*              mask    - array of ints for enable/disable      */
/*				veclen	- length of vector						*/
/*																*/
/*	output:		double median of vector							*/
/****************************************************************/
double median_nvect( double vect[], int mask[], long veclen )
{
	long	i, j, limit;
	long	count = 0;
	double	*tmp_vect;
	double 	median;
	
	if( (tmp_vect = alloc_nvect( veclen )) == NULL )
		return( 0.0 );
	
	/*
	 * copy all enabled cells to temp vector
	 */
	for( i = 0; i < veclen; i++ )
	{
		if( mask[i] )
			tmp_vect[count++] = vect[i];
	}

	limit = (count+1)/2 + 1;
		
	/*
	 * find n/2 largest values
	 * and place at end of list
	 * don't need to sort rest
	 */
	for( i = 0; i < limit; i++, count-- )
	{
		j = maxi_nvect( tmp_vect, display_mask, count );
		SWAPD( tmp_vect[j], tmp_vect[count-1] );
	}
	
	if( EVEN(veclen) )
		median = (vect[limit] + vect[limit+1])/2.0;
	else
		median = vect[limit];
		
	mfree( tmp_vect );
	
	return( median );
}

/****************************************************************/
/*	function:	sort_nvect() - sort values in an n-vector		*/
/*																*/
/*	input:		vect	- n-vector to sort						*/
/*				veclen	- length of vector						*/
/*																*/
/*	output:		none											*/
/****************************************************************/
void sort_nvect( double vect[], long veclen )
{
	long	i, j, limit = veclen;
	
	/*
	 * find n/2 largest values
	 * and place at end of list
	 * don't need to sort rest
	 */
	for( i = 0; i < limit; i++, veclen-- )
	{
		j = maxi_nvect( vect, display_mask, veclen );
		SWAPD( vect[j], vect[veclen-1] );
	}
}

/****************************************************************/
/*	function:	rand_nvect() - generate random vector			*/
/*																*/
/*	input:		vect	- n-vector to fill with random values	*/
/*				veclen	- length of vector						*/
/*																*/
/*	output:		none											*/
/****************************************************************/
void rand_nvect( double vect[], long veclen )
{
	long	i;
	
	for( i = 0; i < veclen; i++ )
		vect[i] = 2.0*(nrand() - 0.5);
}

