/****************************************************************/
/* headers                                                      */
/****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "general.h"
#include "util.h"
#include "viz.h"
#include "nvect.h"
#include "nmatrix.h"
#include "db.h"
#include "cluster.h"
#include "bonds.h"
#include "stats.h"
#include "file.h"
#include "animate.h"
#include "cfg.h"

/****************************************************************/
/* defines                                                      */
/****************************************************************/
#define NOT_FOUND		(-1)
#define SAFE(s)			((s) ? s : "NULL")
#define PRNT_BOOL(n)	((n) ? "TRUE" : "FALSE")

/****************************************************************/
/* typedefs                                                     */
/****************************************************************/

/****************************************************************/
/* external function                                            */
/****************************************************************/

/****************************************************************/
/* function prototypes                                          */
/****************************************************************/
void report_config( DB *db );
int read_config_args( DB *db, int argc, char **argv );
int read_config_file( DB *db, int argc, char **argv );
int parse_config( DB *db );
int param_index( char *name );
int valid_arg( char *arg, int index, int *code );

/****************************************************************/
/* global variables                                             */
/****************************************************************/
char errbuf[400];
int debug = TRUE;
char buf[1024];

void (*error)( char *str );


/****************************************************************/
/*  function:   report_config() - print out configuration of    */
/*              database coordinates to ASCII text file         */
/*                                                              */
/*  input:      db      - pointer to database                   */
/*                                                              */
/*  output:     none                                            */
/****************************************************************/
void report_config( DB *db )
{
	DEBUG( "report_config" );

}


/****************************************************************/
/****************************************************************/

#define LONG_MAX            999999999999999

typedef struct param{
	char	*name;
	int		num_args;
	int		type;
	double	min;
	double	max;
	int		num_choices;
	char	*vals[10];
	int		codes[10];
	}PARAM;

PARAM param_list[] = { \
{"window_width",	 	1,	INTEGER, 	400,	1000,		0			},
{"window_height",	 	1,	INTEGER, 	400,	1000,		0			},
{"graph_width",		 	1,	INTEGER, 	100, 	500,		0			},
{"graph_height",	 	1,	INTEGER, 	100, 	500,		0			},
{"interact_mode",	 	1,	STRING,	   	0,		0, 			4,
										{"rotate","pan","zoom","select" },
										{ ROTATE,  PAN,  ZOOM,  SELECT  }},
{"output_mode",		 	2,	SPECIAL,   	0,		0, 			2,
													 {"text", "graphics"},
													 { TEXT,   GRAPHICS }},
{"draw_mode",		 	10, SPECIAL,   	0,		0, 			10, 
{ "dots", "lines", "trace", "labels", "bonds", "links",  "spheres", 
												"axes", "unit", "bounds" },
{  DOTS,   LINES,   TRACE,   LABELS,   ATTRACTIONS, LINKS, SPHERES, 
												 AXES,   UNIT,   BOUNDS  }},
{"plot_mode",		 	3,	SPECIAL,   	0,		0,			4,
{"none","stress","shepard","correlation"},
{ FALSE, STRESS, SHEPARD, CORRELATION }},
{"output_filename",	 	1, 	STRING,	   	0,		0, 			0			},
{"select_mode",		 	1, 	STRING,	 	0,		0,			2,
												  {"top_ten","random"},
												  { TOP_TEN,  RANDOM_ONCE }},
{"bonds_per_datum",	 	1, 	LONG, 		2, 		LONG_MAX, 	0		},
{"distance_function", 	1,	STRING,	   	0,		0,  		5, 
		{"euclidian", "manhattan", "cosine", "correlation", "minkowski" },
		{ EUCLIDIAN,   MANHATTAN,   COSINE,   CORRELATION,   MINKOWSKI }},
{"stress_function",	 	1,	STRING,	   	0,		0,			7, 
{ "kruskal1", "kruskal2", "lingoes", "sammon", "j_ee", "j_ff", "j_ef" },
{ KRUSKAL_STRESS1, KRUSKAL_STRESS2, LINGOES_STRESS, SAMMON_STRESS, 
  J_EE_STRESS, J_FF_STRESS, J_EF_STRESS }},
{"step_factor",		 	1,	FLOAT,	  	1.0,	3.0,		0			},
{"noise_factor",	 	1,	FLOAT,	  	0.0,	1.0,		0			},
{"gen_screen_coords",	1,	STRING,		0,		0,			3, 
											{"random", "variance", "data"},
											{ RANDOM,   VARIANCE,   DATA }},
{"distance_expression",	1,	STRING,		0,		0,			0			},
{"stress_expression",	1,	STRING,		0,		0,			0			},
{"display_dimensions",	1,	INTEGER,	1,	   	MAX_DISPLAY_DIMS, 0		},
{"viewplane_minimum",	2,	FLOAT,	  	0.0,	0.0,		0			},
{"viewplane_maximum",	2,	FLOAT,	  	0.0,	0.0,		0			},
{"eye_point",		 	3,	FLOAT,	  	0.0,	0.0,		0			},
{"lookat_point",	 	3, 	FLOAT,    	0.0,    0.0,		0			},
{"up_vector",		 	3, 	FLOAT,    	0.0,    0.0,		0			},
{"eye_to_viewplane", 	1, 	FLOAT,    	0.1,    40.0,		0			},	
{"eye_to_front_clip",	1,	FLOAT,	  	0.1,	40.0,		0			},
{"eye_to_back_clip", 	1, 	FLOAT,    	0.1,    40.0,		0			},
{"light_point",      	3, 	FLOAT,    	0.0,    0.0,		0			},
{"radius",				1,	FLOAT,	  	1.0,	40.0,		0			},
{"ambient_rgb",			3,  FLOAT,    	0.0,    1.0,		0			},
{"diffuse_rgb",			3,  FLOAT,    	0.0,    1.0,		0			},
{"specular_rgb",		3,  FLOAT,    	0.0,    1.0,		0			},
{"maximum_iteration",	1,	LONG,		0,	  	LONG_MAX, 	0			},
{"save_frames",			1,  STRING,		0,		0,			8,
		{ "yes", "no", "true", "false", "on", "off", "enabled", "disabled" },
		{  TRUE, FALSE, TRUE,   FALSE,  TRUE, FALSE,  TRUE,		FALSE   }},
{"frame_base_filename",	1,	STRING,		0,		0,			0			},
{"first_frame_saved",	1,	LONG,		0,	    LONG_MAX, 	0			},
{"last_frame_saved",	1,	LONG,		0,		LONG_MAX, 	0			},
{"skip_frames",			1,  INTEGER,	0,		LONG_MAX, 	0			},
{"data_source",			1,  STRING,		0,		0, 			8,
{"no_data","from_file","sim_tri","sim_cube","sim_hypercube","sim_cluster",
 "sim_tetra", "sim_tetrahedron" }, 
{ NO_DATA,  FROM_FILE,  SIM_TRI,  SIM_CUBE,  SIM_HYPERCUBE,  SIM_CLUSTER,
  SIM_TETRAHEDRON, SIM_TETRAHEDRON }}, 
{"data_filename",		1,	STRING,		0,		0,			0			},
{"normalize_function", 	1,	STRING,	   	0,		0,  		4, 
					{"none",   "scale", "subtract_mean", "unit_variance" },
					{ NO_NORM, SCALE,    SUB_MEAN,        UNIT_VARIANCE }},
{"interpolate",			1,  STRING,		0,		0,			8,
		{ "yes", "no", "true", "false", "on", "off", "enabled", "disabled" },
		{  TRUE, FALSE, TRUE,   FALSE,  TRUE, FALSE,  TRUE,		FALSE   }},
{ "num_interpolate_steps", 1, INTEGER, 0, 		10,			0			},
{ "back_color",			3,	INTEGER,  	0,		255,		0			},
{ "axis_color",			3,	INTEGER,  	0,		255,		0			},
{ "low_stress_color",	3,	INTEGER,  	0,		255,		0			},
{ "high_stress_color",	3,	INTEGER,  	0,		255,		0			},
{ "cluster_probability",1,  FLOAT,    	0.0,    1.0,		0			},
{ "num_data_points",	1,  LONG,    	1,    	LONG_MAX,	0			},
{ "num_data_dims",		1,  LONG,    	1,    	LONG_MAX,	0			},
{"",					0,	0,			0,		0,			0			},
};

int read_config_args( DB *db, int argc, char **argv )
{
	int		stat;

	DEBUG( "read_config_args" );

	reset_tokens( TOKEN_CMDLINE, NULL, argc, argv );
	stat = parse_config( db );

	return( stat );
}

int read_config_file( DB *db, int argc, char **argv )
{
	FILE	*fp;
	int		stat;

	DEBUG( "read_config_file" );

	if( (fp = fopen("mds.cfg", "r")) == NULL )
		return( ERROR );

	reset_tokens( TOKEN_FILE, fp, 0, NULL );
	stat = parse_config( db );

	fclose( fp );

	return( stat );
}

int parse_config( DB *db )
{
	char	tok[200];
	char	eq[10];
	char	val[10][200];
	int		i, index;
	int		code[10];
	double	dval[10];
	double	min, max;
	char	*end;
	char	*ptr;
	int		output_mode = 0;
	int		draw_mode = 0;
	int		plot_mode = 0;
	int		num_special_args;

	DEBUG( "parse_config" );

	while( get_token(tok,200) )
	{
		/*
		 * if parameter is specified without spaces (i.e. name=value)
		 * split parameter into parts and put '=' and "value" onto
		 * input buffer so they will be read off by next get_token()
		 */
		if( (ptr = strchr(tok,'=')) != NULL )
		{
			
			*ptr = '\0';
			unget_token( ++ptr );
			unget_token( "=" );
		}

		/*
		 * see if token is valid parameter
		 */
		if( (index = param_index(tok)) == NOT_FOUND )
		{
			sprintf( buf, "Unrecognized program parameter: %s", tok );
			print_error( buf );
			return( ERROR );
		}

		/*
		 * check for required '=' sign
		 */
		if( !get_token(eq,10) || strcmp( eq, "=") != 0 )
		{
			sprintf( buf, "missing '=' in config parameter: %s", tok );
			print_error( buf );
			return( ERROR );
		}

		/*
		 * read in number of arguments appropriate for this parameter
		 */
		for( i = 0; i < param_list[index].num_args; i++ )
		{
			if( !get_token(val[i], 200) )
			{
				if( i > 0 && param_list[index].type == SPECIAL )
					break;
				else
				{
					sprintf(buf,"missing argument for config parameter:%s",tok);
					print_error( buf );
					return( ERROR );
				}
			}

			/*
			 * if the parameter has a comma in it, put
			 * whatever's after the comma back on the
			 * input stream. If we have read in a comma
			 * on its own just discard it and reread the
			 * next argument
			 */
			if( (ptr = strchr(val[i],',')) != NULL )
			{
				*ptr = '\0';

				if( strlen(ptr+1) > 0 )
					unget_token( ++ptr );

				if( strlen(val[i]) == 0 )
				{
					i--;
					continue;
				}
			}

			/*
			 * the "special" parameters can have a variable
			 * number of arguments, so we read parameters
			 * until we get an illegal one, and then put
			 * that back on the input stream
			 */
			if( param_list[index].type == SPECIAL && 
							!valid_arg(val[i],index,&code[i]) )
			{
				unget_token( val[i] );
				break;
			}
		}

		if( param_list[index].type == SPECIAL )
			num_special_args = i;

		/*
		 * if parameter is a string check it against
		 * the list of legal string entries
		 */
		if( param_list[index].type == STRING && 
			!valid_arg(val[0], index, &code[0]) )
		{
			sprintf(buf,"invalid argument parameter for %s: %s",tok,val[0]);
			print_error( buf );
			return( ERROR );
		}

		/*
		 * if the parameter is a number convert it to a double
		 * and if the table specifies a max and a min, clamp
		 * value between max and min
		 */
		if( param_list[index].type == INTEGER ||
			param_list[index].type == LONG ||
			param_list[index].type == FLOAT )
		{
			min = param_list[index].min;
			max = param_list[index].max;

			for( i = 0; i < param_list[index].num_args; i++ )
			{
				/*
				 * get value in double format
				 */
				dval[i] = strtod( val[i], &end );

				/*
				 * if min and max are not both set to zero, CLAMP
				 * values between min and max
				 */
				if(	!(FLT_ZERO(min) && FLT_ZERO(max)) )
					dval[i] = CLAMP( dval[i], min, max);
			}
		}

		/*
		 * store parameter in appropriate 
		 * variable(s) in database
		 */
		switch( index )
		{
			case 0:	
//					db->view.params[SCREEN_MAX][X] = (int)dval[0];
					break;
			case 1:	
//					db->view.params[SCREEN_MAX][Y] = (int)dval[0];
					break;
			case 2:
//					db->shepard.width = (int)dval[0];
//					db->stress.width = (int)dval[0];
//					db->correlate.width = (int)dval[0];
					break;
			case 3:
//					db->shepard.height = (int)dval[0];
//					db->stress.height = (int)dval[0];
//					db->correlate.height = (int)dval[0];
					break;
			case 4:
					db->interact_mode = code[0];
					break;
			case 5:
					for( i = 0; i < num_special_args; i++ )
						output_mode |= code[i];
					break;	
			case 6:
					for( i = 0; i < num_special_args; i++ )
						draw_mode |= code[i];
					break;	
			case 7:
					for( i = 0; i < num_special_args; i++ )
						plot_mode |= code[i];
					break;	
			case 8:
					if( strcmp(strip_quotes(val[0]), "NULL") != 0 )
						new_str( &db->files.outfile, val[0] );
					break;	
			case 9:
					db->mds.select_mode = code[0];
					break;	
			case 10:
					db->mds.num_bonds = (long)dval[0];
					db->mds.num_bonds += ODD(db->mds.num_bonds) ? 1: 0;
					break;	
			case 11:
					db->mds.dist_func = code[0];
					break;
			case 12:
					db->mds.stress_function = code[0];
					break;
			case 13:
					db->mds.step_factor = dval[0];
					break;
			case 14:
					db->mds.noise_factor = dval[0];
					break;
			case 15:
					db->mds.display_init_mode = code[0];
					break;
			case 16:
					if( strcmp(strip_quotes(val[0]), "NULL") != 0 )
						new_str( &db->mds.dist_expression, val[0]);
					break;
			case 17:
					if( strcmp(strip_quotes(val[0]), "NULL") != 0 )
						new_str( &db->mds.stress_expression, val[0]);
					break;
			case 18:
					db->dat.display_dims = (int)dval[0];
					break;
			case 19:
//					set_vect( db->view.params[VIEW_MIN],dval[0],dval[1],0,1);
					break;
			case 20:
//					set_vect( db->view.params[VIEW_MAX],dval[0],dval[1],0,1);
					break;
			case 21:
//					set_vect( db->view.params[EYE], dval[0],dval[1],dval[2],1);
//					copy_vect( db->view.params[OLD_EYE], db->view.params[EYE] );
					break;
			case 22:
//				   set_vect( db->view.params[LOOK_AT],
//							 dval[0], dval[1], dval[2], 1 );
					break;
			case 23:
//					set_vect(db->view.params[UP],dval[0],dval[1],dval[2],1);
					break;
			case 24:
//					set_vect(db->view.params[VIEW_DIST], dval[0], 0, 0, 1 );
					break;
			case 25:
//					set_vect(db->view.params[FRONT], dval[0], 0, 0, 1 );
					break;
			case 26:
//					set_vect(db->view.params[BACK], dval[0], 0, 0, 1 );
					break;
			case 27:
//				  set_vect( db->view.params[LIGHT_PT],
//							dval[0], dval[1], dval[2],1);
					break;
			case 28:
//					db->spheres.radius = dval[0];
					break;
			case 29:
//					set_vect( db->spheres.properties.ambient,
//							  dval[0], dval[1], dval[2], 1 );
					break;
			case 30:
//					set_vect( db->spheres.properties.diffuse,
//							  dval[0], dval[1], dval[2], 1 );
					break;
			case 31:
//					set_vect( db->spheres.properties.specular,
//								dval[0], dval[1], dval[2], 1 );
					break;
			case 32:
					db->animate.max_iteration = (long)dval[0];
					break;
			case 33:
					db->animate.save_frames = code[0];
					break;
			case 34:
					if( strcmp(strip_quotes(val[0]), "NULL") != 0 )
						new_str( &db->animate.frame_base, val[0] );
					break;
			case 35:
					db->animate.start_save_iter = (long)dval[0];
					break;
			case 36:
					db->animate.end_save_iter = (long)dval[0];
					break;
			case 37:
					db->animate.save_skip = (int)dval[0];
					break;
			case 38:
					db->dat.data_init_mode = code[0];
					break;
			case 39:
					if( strcmp(strip_quotes(val[0]), "NULL") != 0 )
						new_str( &db->files.infile, val[0] );
					break;	
			case 40:
					db->mds.norm_function = code[0];
					set_normalize( db->mds.norm_function );
					break;
			case 41:
					db->animate.interpolate = code[0];
					break;
			case 42:
					db->animate.interp_steps = (int)dval[0];
					break;
			case 43:
//					init_private_color( db, BACK_COLOR, 
//										dval[0], dval[1], dval[2] );
					break;

			case 44:
//				   init_private_color( db, AXIS_COLOR,
//									   dval[0], dval[1], dval[2] );
					break;
			case 45:
//					init_private_color( db, LOW_STRESS_COLOR,
//					 				    dval[0], dval[1], dval[2] );
					break;
			case 46:
//					init_private_color( db, HIGH_STRESS_COLOR,
//									    dval[0], dval[1], dval[2] );
					break;
			case 47:
					db->dat.clust_prob = dval[0];
					break;
			case 48:
					db->dat.num_points = (long) dval[0];
					break;
			case 49:
					db->dat.data_dims = (long) dval[0];
					break;
			default:
					break;
		}
	}

	if( output_mode != 0 )
		db->output_mode = output_mode;
	
	if( draw_mode != 0 )
		db->draw_mode = draw_mode;
	
	if( plot_mode != 0 )
		db->plot_mode = plot_mode;

	return( SUCCESS );
}

int param_index( char *name )
{
	char	low_name[200];
	int		i;

	DEBUG( "param_index" );

	strcpy( low_name, name );
	tolower_string( low_name );

	for( i = 0; param_list[i].name[0] != '\0'; i++ )
	{
		if( streq(low_name, param_list[i].name) )
			return( i );
	}

	return( NOT_FOUND );
}

int valid_arg( char *arg, int index, int *code )
{
	int		choices = param_list[index].num_choices;
	char	low_arg[200];
	int		i;

	DEBUG( "valid_arg" );

	if( choices == 0 )
		return( TRUE );

	strcpy( low_arg, arg );
	tolower_string( low_arg );

	for( i = 0; i < choices; i++ )
	{
		if( streq(low_arg, param_list[index].vals[i]) )
		{
			*code = param_list[index].codes[i];
			return( TRUE );
		}
	}

	return( FALSE );
}

