#ifndef DocumentManager_H
#define DocumentManager_H

#include <string>
#include <fstream>
#include "basic/Iterator.h"
#include "basic/macros.h"
#include "DataStructure/DimensionSpace.h"
#include "DataStructure/Query.h"

class ViewManager;
class MainFrame;
class DBManager;
class FlatDataSet;
class HierDataSet;
class VisualElementMeta;

//name code:  dim - dimension, db - database, num - number, doc - document, item - data item (rows in spreadsheet), hash - hash table
class DocumentManager
{
protected:
    ////the following can be accessed through public functions of this class
    ///the following variables are general, i.e., they exist no matter if a dataset is opened or not
    //pointer to the view manager
    ViewManager* viewManager;
    //pointer to the main interface
    MainFrame* mainFrame;
    //the database manager is created and deleted by the doc manager
    DBManager* dbManager;
    
    //this variable defines how to select dimensions
    int dimSelectMode;
    //this variable defines how to display images in the image view
    int imageDisplayMode;
    //this variable defines how to show dimensions in the var view
    int varDisplayMode;
    //this variable decides how to select distince dimension - for less cluttered labeling or for better grouping
    int dimDistinctSelectOption;

    ///the following variables are related to the dataset currently opened. 
    ///If no dataset is opened, they are NULL. 
    //the dim space composed all dimensions in the dataset
    DimensionSpace* allDimSpace;
    //a sub space of the all dim space composed of selected dimensions
    //to do: multiple selected dim spaces
    DimensionSpace* selectedDimSpace;
    //the dim space currently used by the drawers. It can be either the selected dim space or the all dim space
    DimensionSpace* currDimSpace;
    //if this variable is true, then the curr dim space is the selected dim space. Otherwise it is the all dim space
    bool useSelectedDimSpace;
    
    //the hash table that stores the positions of dimensions in a (0, 0) to (1, 1) 2D space.
    IPHash dimXYPosHash;
    //the hash table that stores the positions of data items in a (0, 0) to (1, 1) 2D space.
    IPHash itemXYPosHash;
    //the hash table that stores ids of the selected dimensions
    IIHash selectedDimHash;
    //the hash table that stores ids of the selected data items
    IIHash selectedItemHash;
    //the hash table that stores ids of the selected bins
    IIHash selectedBinHash;
    //pointer to the hash table that stores ids of the labeled dimensions
    //notice that this is a pointer - it can point to the selected dim hash or the manual label dim hash
    //when it points to selected dim hash, all selected dimensions are labeled. Otherwise it is not true. 
    IIHash* labelDimHash;
    //the hash table that stores ids of the dimensions user manuall selected for labeling
    IIHash manualLabelDimHash;

    ////the following varibles' values can be changed or read through public functions of this class
    ///the following variables are general, i.e., they exist no matter if a dataset is opened or not
    //dbMode is either true (database is connected) or false  
    bool dbMode;

    ///the following variables are related to the dataset currently opened. 
    ///If no dataset is opened, they are empty (hashtables), -1(integers) or ""(strings).
    std::string dataSetName;
    
    Query* query;

    //variable related to the var control frame. See varcontrolframe.h for detail
    int activeItemNum;
    int activeDimID;
    int baseDimID;
    //time control for the rainfall animation in the var display
    int centerDimID;
    float centerT;

    //variable related to the MDS control frame. See mdscontrolframe.h for detail
    int focusItemID;
    int baseItemID;
    //for the rainfall animation in the image display
    float centerItemT;
    std::string globalPath;

//the following variable is used within this class only 
private:
    ///the following variables are related to the dataset currently opened. If no dataset is opened, they are NULL.
    FlatDataSet* flatDataSet;
    HierDataSet* hierDataSet;
    int dimNum;
    int itemNum;
    bool hasDescription;
    //the correlation matrix of the dimensions 
    double** cov_matrix;
    //the correlations between the base dimensoins and all dimensions (used for dimension selection)
    float* distancesToBaseDim;
    //the correlations between the base data item and all data items (used for selection by sample)
    float* distancesToBaseItem;
    //the correlations between the center data item and all data items (used for the image rainfall display)
    float* distancesToCenterItem;
    //the random horizontal positions of dimensions in the VaR rainfall display
    float* randomX;
    //the random horizontal positions of images in the image rainfall display
    float* randomItemX;

    //for animated MDS
    bool mds_done;
    bool mds_inited;

    int preImageDisplayMode;
    int preVarDisplayMode;
    int preOrderbyDimID;

    //plan to remove this variable
    bool binaryMode;

public:
    DocumentManager();
	~DocumentManager();

    //set and get pointers to view managers and main interface
    void SetViewManager(ViewManager* viewManager) {this->viewManager = viewManager;}
    ViewManager* GetViewManager() {return viewManager;}
    void SetMainFrame(MainFrame* mainFrame) {this->mainFrame = mainFrame;}
    MainFrame* GetMainFrame() {return mainFrame;}

    bool SetDBMode(bool db);  //db = true, DBMode, db = false, InMemory mode
    DBManager* GetDBManager();
    bool IsDBMode();

    //close the currently opened dataset
    void CloseDataSet();
    //open a new dataset
    bool OpenDataSet(std::string fileName);

    //return true if no dataset is opened or there is no dimensions in the curr dim space 
    bool IsEmpty();

    std::string GetGlobalPath() {return globalPath;}
    void SetGlobalPath(std::string path);


    //functions for getting data
    Iterator* GetElementsIterator(int layerType, VisualElementMeta* meta);
    Iterator* GetFlatDataIterator(DimensionSpace* loadSpace, bool dimCluster, bool brushed, int outputType, bool loadDescriptions, bool storeDescriptions);
    Iterator* GetBinIterator(bool dimCluster, bool brushed);
    Iterator* GetTimeHistogramIterator(DimensionSpace* loadSpace);
    Iterator* GetHierDataIterator(bool dimCluster, bool showBand, bool brushed);
    Iterator* GetColumnIterator(DimensionSpace* loadSpace);
    Iterator* GetAxesIterator(bool dimXYNeeded, bool itemXYNeeded, bool binXYNeeded);
    Iterator* GetDimensionNameIterator();
    Iterator* GetDimensionLabelIterator();
    Iterator* GetItemNameIterator();


    //functions related to dimension space or individual dimension
    DimensionSpace* GetCurrentDimensionSpace();
    DimensionSpace* GetAllDimensionSpace() {return allDimSpace;}
    Dimension* GetDimensionByID(int dimID);
    void SetCurrentDimensionSpace(DimensionSpace* dimSpace);
    bool SelectDimensionsAroundBase(float thres);
    void SetDimensionSelectMode(int i);
    void SetUseSelectedDimensionSpace(bool use);
    void SetDistinctDimensionSelectOption(int i);
    //the three functions are called by the function: SelectDimensionsAroundBase
    bool SelectDistinctDimensionsAroundBaseForLabeling(int seed_dim, float threshold);
    bool SelectDistinctDimensionsAroundBaseForGroups(int seed_dim, float threshold);
    bool SelectSimilarDimensionsAroundBase(int dimID, float thres);

    //functions handle parameters related to var control frame
    bool SetActiveDimension(std::string dimName);
    int GetActiveDimensionID() {return activeDimID;}

    Query* GetQuery() {return query;}

    bool SetOrderbyDimension(int dimID) {query->SetOrderByDimensionID(dimID); return true;}
    bool SetOrderbyDimension(std::string dimName);

    bool SetBaseDimension(std::string dimName, int &newID);
    int GetBaseDimensionID() {return baseDimID;}

    bool SetVarGlyphBaseDimension(std::string dimName, int mode);

    bool SetGroupDimension(std::string dimName);
    void SetGroupValue(std::string value) {query->SetGroupValue(value);}

    bool SetXDimension(std::string dimName);
    bool SetYDimension(std::string dimName);

    //for rainfall var display. centerdimension is the dimension on the botton center, centert is the time of the rain
    bool SetCenterDimension(std::string dimName);
    int GetCenterDimensionID() {return centerDimID;}
    void SetCenterT(float t) {centerT = t;}

    //functions handle parameters related to image control frame
    bool SetFocusItem(std::string itemName, int &newID);
    int GetFocusItemID() {return focusItemID;}

    //for rainfall image display. center item is the image the botton center, centert is the time of the rain
    void SetCenterItemT(float t) {centerItemT = t;}

    bool SetBaseItem(std::string itemName, int &newID);
    int GetBaseItemID() {return baseItemID;}
    void GetDistancesFromBaseItem();

    bool SelectItemsAroundBase(float thres);
    bool SelectSelectedItemByDimensionValue(int dimID, float value1, float value2);
    bool SelectItemByDimensionValue(int dimID, float value1, float value2);

    void ResetVarLabelingMode(int i);
    void SetSamplingMode(int i);

    //functions for image positioning in image display
    void ResetImagePositions(int i);
    void CreateMDSPosForItems();
    bool OnlineCreateOneByOneSelectedItemPos();
    bool OnlineCreateMDSSelectedItemPos();
    bool OnlineCreateAroundCenterItemPos();

    //functions for dimension positioning in var display
    void ResetVarPositions(int i);
    bool CreateMDSPosForDimensions();
//    bool CreateAnimateMDSPos();
    bool OnlineCreateAroundCenterPos();
    bool OnlineCreateMDSSelectedDimensionPos();
    bool OnlineCreateMDSSelectedDimensionPos2();
    bool OnlineCreateXYDimensionPos();
    bool CreateCurveTreemapPosForDimensions();

    //to be removed
    bool IsBinaryMode() {return binaryMode;}
    void SetBinaryMode(int i);
    
    Dimension* GetTimeDimension();

    int GetPrevImageDisplayMode() {return preImageDisplayMode;}

    IIHash* GetSelectedItemHash() {return &selectedItemHash;}

    void SetVarGlyphMode(int i);
    
    std::string GetItemDescription(int itemID);


protected:
    bool ReadFlatDataSet(std::string fileName);
    bool ReadCgFile(std::string fileName);
    bool ReadMetaData(std::ifstream* inFile); 
    bool ReadHierData(std::string fileName); 


private:

    void EmptyDimXYPosHash();
    void EmptyItemXYPosHash();
    void GetSelectedDimensionSpaceFromSelectedDimensionHash();

    int GetDimensionCovMatrix();
    void GetDistancesFromCenterItem();
    void GetDistancesFromBaseDimension();
    int GetItemIDByName(std::string itemName);
    int GetDimensionIDByName(std::string itemName);

    void LoadGroupInformation(std::string fileName);

};


#endif

