/****************************************************************/
/* headers                                                      */
/****************************************************************/
#include <stdio.h>
#include <malloc.h>
#include "general.h"
#include "util.h"

/****************************************************************/
/* defines                                                      */
/****************************************************************/

/****************************************************************/
/* typedefs                                                     */
/****************************************************************/

/****************************************************************/
/* external function                                            */
/****************************************************************/

/****************************************************************/
/* function prototypes                                          */
/****************************************************************/
void error_tty( char *str );
void error_X( char *str );
char *mstrdup( char * );
void tolower_string( char *str );
void toupper_string( char *str );
void new_str( char **str_ptr, char *str );
void reset_tokens( int mode, FILE *fp, int argc, char **argv );
int get_token( char *tok, int size );
int unget_token( char *tok );
int read_line( FILE *fp, char *buf, int size );
char *strip_quotes( char *str );
#ifdef MY_MALLOC
char *mmalloc( UINT size );
char *mcalloc( UINT nelem, UINT elsize );
char *mrealloc( void *oldptr, UINT size );
void mfree( void *ptr );
UINT get_mem();
#endif

/****************************************************************/
/* global variables                                             */
/****************************************************************/
void (*error)( char *str );
static ULONG mem = 0;
static char buf[400] = { '\0' };
static char *ptr = &buf[0];
static int tok_mode = TOKEN_FILE;
static int tok_argc = 0;
static int cur_tok = 1;
static char **tok_argv = NULL;
static FILE *tok_fp = NULL;
 
/****************************************************************/
/* code															*/
/****************************************************************/
void error_tty( char *str )
{
	if( str )
		fprintf( stderr, "%s\n", str );
}

/************************************************************************/
/*  Function:   mstrdup - allocates memory to make a copy of a string   */
/*                                                                      */
/*  Inputs:                                                             */
/*      str - pointer to string to copy                                 */
/*                                                                      */
/*  Returns:                                                            */
/*      NULL if there is no input string, or the malloc fails,          */
/*      otherwise it returns a pointer to the newly allocated string    */
/************************************************************************/
char *mstrdup( char *str )
{
    char    *tmp;

    if( !str )
        return( NULL );

    if( (tmp = (char *)mmalloc((strlen(str)+1)*sizeof(char))) == NULL )
        return( NULL );

    strcpy( tmp, str );

    return( tmp );
}

/************************************************************************/
/*  Function:   tolower_string - changes all the characters in string   */
/*              to lower case by calling tolower().                     */
/*                                                                      */
/*  Inputs:                                                             */
/*      str - pointer to characters to convert                          */
/*                                                                      */
/*  Returns:                                                            */
/*      none                                                            */
/************************************************************************/
void tolower_string( char *str )
{
    for( ; *str != '\0'; str++ )
        *str = tolower(*str);
}

/************************************************************************/
/*  Function:   toupper_string - changes all the characters in string   */
/*              to upper case by calling toupper().                     */
/*                                                                      */
/*  Inputs:                                                             */
/*      str - pointer to characters to convert                          */
/*                                                                      */
/*  Returns:                                                            */
/*      none                                                            */
/************************************************************************/
void toupper_string( char *str )
{
    for( ; *str != '\0'; str++ )
        *str = toupper(*str);
}

/************************************************************************/
/*  Function:   new_str - frees a string and allocates a new one		*/
/*				calls mstrdup to allocate the new string, so if			*/
/*				mstrdup fails the character pointer may be set to NULL	*/

/************************************************************************/
/*  Function:   new_str - frees a string and allocates a new one		*/
/*				calls mstrdup to allocate the new string, so if			*/
/*				mstrdup fails the character pointer may be set to NULL	*/
/*                                                                      */
/*  Inputs:                                                             */
/*      str_ptr - address of character pointer to point at new string	*/
/*      str 	- pointer to string to copy                             */
/*                                                                      */
/*  Returns:                                                            */
/*		none															*/
/************************************************************************/
void new_str( char **str_ptr, char *str )
{
    if( *str_ptr )
        mfree( *str_ptr );

    *str_ptr = mstrdup( str );
}

/************************************************************************/
/* reset_tokens															*/
/************************************************************************/
void reset_tokens( int mode, FILE *fp, int argc, char **argv )
{
	buf[0] = '\0';
	ptr = &buf[0];

	tok_mode = CLAMP(mode, TOKEN_FILE, TOKEN_CMDLINE );
	cur_tok = 1;
	tok_argc = argc;
	tok_argv = argv;
	tok_fp = fp;
}

/************************************************************************/
/* unget_token															*/
/************************************************************************/
int unget_token( char *tok )
{
	int		len, buflen;
	char	tmp[400];
	char	*ptr2;

	if( !tok )
		return( ERROR );

	/*
	 * get length of token we're adding back onto
	 * buffer + length of unparsed portion of buffer
	 */
	len = strlen( tok );
	buflen = strlen( ptr );

	/*
	 * silently ignore a token with no length
	 */
	if( len == 0 )
		return( SUCCESS );

	/*
	 * check if buffer too small
	 */
	if( (len + buflen + 1) >= 400 )
		return( ERROR );

	/*
	 * copy new tok and rest of buffer
	 * back to start of buffer and reset
	 * pointer to start of buffer
	 */
	sprintf( tmp, "%s %s", tok, ptr );
	strcpy( buf, tmp );

	ptr = &buf[0];

	return( SUCCESS );
}

/************************************************************************/
/* get_token															*/
/************************************************************************/
int get_token( char *tok, int size )
{
	int		len;
	char	*ptr2;

	while( *ptr == '\0' )
	{
		/*
		 * read in line in one of several modes
		 */
		switch( tok_mode )
		{
			case TOKEN_FILE:
				if( tok_fp && !fgets(buf, 400, tok_fp) )
					return( FALSE ); 
				break;

			case TOKEN_CMDLINE:
				if(	!tok_argv || !tok_argv[cur_tok] || cur_tok == tok_argc )
					return( FALSE );

				/*
				 * copy argument to buffer as if it were line
				 * and add a newline and NULL to be like fgets()
				 */
				strncpy( buf, tok_argv[cur_tok++], 398 );
				len = strlen( buf );
				buf[len++] = '\n';
				buf[len++] = '\0';
				break;

			default:
				return( FALSE );
				break;
		}

		/*
		 * skip blank lines and lines with '#' at the start
		 */
		if( buf[0] == '#' || buf[0] == '\n' )
		{
			*ptr = '\0';
			continue;
		}

		/*
		 * get rid of '\n' at end of line
		 */
		buf[strlen(buf)-1] = '\0';
		
		/*
		 * cut off to the right of comments
		 */
		if( ptr2 = strchr(buf, '#') )
			*ptr2 = '\0';

		/*
		 * set ptr to start of new line
		 */
		ptr = &buf[0];

		/*
		 * find first non-space
		 */
		while( isspace(*ptr) )
			ptr++; 
	}

	ptr2 = ptr;

	/*
	 * find first space after token
	 */
	while( !isspace(*ptr2) && *ptr2 != '\0' )
		ptr2++;

	len = MIN((ptr2 - ptr), size);

	/*
	 * if we've hit last token
	 */
	if( *ptr2 == '\0' )
	{
		strncpy( tok, ptr, len );
		tok[len] = '\0';
		ptr = ptr2;
	}
	else
	{
		*ptr2 = '\0';

		strncpy( tok, ptr, len );
		tok[len] = '\0';

		ptr = ++ptr2;

		/*
		 * prepare for next invocation by
		 * finding first non-space - if this
		 * sets ptr to point to the NULL character
		 * at the end of the line, then the next
		 * call to get_token will trigger a new
		 * line to be read in
		 */
		while( isspace(*ptr) && *ptr != '\0' )
			ptr++;
	}

	return( TRUE );
}

/************************************************************************/
/*  Function:   read_line - reads in and discards lines that are        */
/*              empty or start with a '#' character, or have only       */
/*              whitespace before the '#' character.  When a            */
/*              non-comment line is encountered it is returned to the   */
/*              calling function                                        */
/*                                                                      */
/*  Inputs:                                                             */
/*      fp  - FILE pointer to read data from                            */
/*      buf - character array to copy line to                           */
/*      int - integer size of character array                           */
/*                                                                      */
/*  Returns:                                                            */
/*      ERROR if the end-of-file is encountered, SUCCESS otherwise      */
/************************************************************************/
int read_line( FILE *fp, char *buf, int size )
{
    char    *ptr;

    /*
     * read lines - return ERROR if read fails
     */
    while( fgets( buf, size, fp ) )
    {
        /*
         * skip blank lines and comments
         */
        if( buf[0] == '#' || buf[0] == '\n' )
            continue;

       /*
         * if no comment character return line
         */
        if( (ptr = strchr( buf, '#' )) == NULL )
            return( SUCCESS );

        /*
         * discard line beyond comment
         */
        *ptr = '\0';

        /*
         * check if remaining line is
         * all whitespace - if so keep
         * reading lines
         */
        for( ptr = buf; *ptr; ptr++ )
        {
            if( !isspace(*ptr) )
			{
                strcpy( buf, ptr );
                return( SUCCESS );
            }
        }
    }

    return( ERROR );
}

char *strip_quotes( char *str )
{
	char	*from = str;
	char	*to = str;

	while( *from )
	{
		if( *from == '"' )
			from++;
		else
			*to++ = *from++;
	}

	*to = '\0';

    return( str );
}

#ifdef MY_MALLOC

char *mmalloc( UINT size )
{
	char	*ptr;

	/*
	 * allocate memory + 4 extra bytes at the start
	 * and 4 extra bytes at the end
	 */
	if( (ptr = (char *)malloc(size+(2*sizeof(UINT)))) == NULL )
	{
		print_error( "Memory allocation error" );
		return( NULL );
	}

	/*
	 * store the size in the 4 bytes before the
	 * start of the returned memory. Also save
	 * the size in the 4 bytes at the end as a
	 * sort of checksum to detect memory overruns
	 */
	memcpy( ptr, &size, sizeof(UINT) );
	memcpy( ptr+size+sizeof(UINT), &size, sizeof(UINT) );
	/*
	*((UINT *)ptr) = size;
	*((UINT *)(ptr+size+sizeof(UINT))) = size;
	*/

	/*
	 * update total count of memory usage
	 */
	mem += size;

	return( ptr+sizeof(UINT) );
}

char *mcalloc( UINT nelem, UINT elsize ) 
{
	char	*ptr;
	UINT	size;

	size = nelem * elsize;

	/*
	 * allocate memory + 4 extra bytes at the start
	 * and 4 extra bytes at the end
	 */
	if( (ptr = mmalloc(size)) == NULL )
		return( NULL );

	memset( ptr, 0, size );

	return( ptr );
}

char *mrealloc( void *oldptr, UINT size )
{
	UINT	oldsize;
	UINT	oldsize2;
	char	*ptr;
	char	*oldcptr;

	/*
	 * decrement pointer to memory to start of 
	 * 4 bytes which preceeds it and retrieve
	 * size that is stored there. Also retrieve
	 * the size from the 4 bytes at the end as a
	 * sort of checksum to detect memory overruns
	 */
	oldcptr = (char *)oldptr;
	oldcptr -= sizeof(UINT);

	memcpy( ptr, &oldsize, sizeof(UINT) );
	memcpy( oldcptr+oldsize+sizeof(UINT), &oldsize2, sizeof(UINT) );
	
	/*
	oldsize = *((UINT *)oldcptr);
	oldsize2 = *((UINT *)(oldpctr+oldsize+sizeof(UINT)));
	*/

	/*
	 * compare two copies of size to see if we ran over
	 * the end of this memory segment
	 */
	if( oldsize != oldsize2 )
		print_error( "Memory corruption found in mrealloc()" );

	/*
	 * decrement total memory usage counter
	 */
	mem -= oldsize;

	if( (ptr = realloc(oldcptr, size+(2*sizeof(UINT)))) == NULL )
	{
		print_error( "Memory re-allocation error" );
		return( NULL );
	}

	/*
	 * store the size in the 4 bytes before the
	 * start of the returned memory. Also save
	 * the size in the 4 bytes at the end as a
	 * sort of checksum to detect memory overruns
	 */
	memcpy( ptr, &size, sizeof(UINT) );
	memcpy( ptr+size+sizeof(UINT), &size, sizeof(UINT) );
	/*
	*((UINT *)ptr) = size;
	*((UINT *)(ptr+size+sizeof(UINT))) = size;
	*/

	/*
	 * update total count of memory usage
	 */
	mem += size;

	return( ptr+sizeof(UINT) );
}

void mfree( void *ptr )
{
	char	*cptr;
	UINT	size;
	UINT	size2;

	if( ptr == NULL )
		return;

	/*
	 * decrement pointer to memory to start of 
	 * 4 bytes which preceeds it and retrieve
	 * size that is stored there. Also retrieve
	 * the size from the 4 bytes at the end as a
	 * sort of checksum to detect memory overruns
	 */
	cptr = (char *)ptr;
	cptr -= sizeof(UINT);

	memcpy( &size, cptr, sizeof(UINT) );
	memcpy( &size2, cptr+size+sizeof(UINT), sizeof(UINT) );
	/*
	size = *((UINT *)cptr);
	size2 = *((UINT *)(cptr+size+sizeof(UINT)));
	*/

	/*
	 * compare two copies of size to see if we ran over
	 * the end of this memory segment
	 */
	if( size != size2 )
		print_error( "Memory corruption found in mfree()" );

	/*
	 * decrement total memory usage counter
	 */
	mem -= size;

	/*
	 * free memory
	 */
	free( (void *)cptr );
}

UINT get_mem()
{
	return( mem );
}

#endif
